# -*- coding: utf-8 -*-
r"""
This subpackage of the spinterface includes python implementations of algorithm which final versions may or may not be
implemented in the spin dynamic program. They can represent straightforward implementations without optimization written
in easy readable python code or test cases for low dimensionalities. These algorithms are also sometimes the blueprint
for a serious fortran implementation. Also they can be used to demonstrate the way the algorithms work. In this context
the sub-package visualizations.energylandscape is of interest since with that it's possible to visualize the behaviour
of these algorithms on a low dimensional energy surface. So far this package contains the following algorithms:

========================================================================================================================



"""