# -*- coding: utf-8 -*-
r"""
contains constants for visualization
"""
SPINDEFAULT_TIPLENGTH: float = 1.0
SPINDEFAULT_TIPRADIUS: float = 0.35
SPINDEFAULT_ARROWSCALE: float = 0.7
SPINDEFAULT_DRAWBACKGROUND: bool = False
SPINDEFAULT_SETTINGS={'tiplength':SPINDEFAULT_TIPLENGTH,'tipradius':SPINDEFAULT_TIPRADIUS,'arrowscale':SPINDEFAULT_ARROWSCALE, 'drawbackground':SPINDEFAULT_DRAWBACKGROUND}

EVECDEFAULT_TIPLENGTH: float = 0.4
EVECDEFAULT_TIPRADIUS: float = 0.25
EVECDEFAULT_ARROWSCALE: float = 5.0
EVECDEFAULT_DRAWBACKGROUND: bool = True
EVECDEFAULT_SETTINGS={'tiplength':EVECDEFAULT_TIPLENGTH,'tipradius':EVECDEFAULT_TIPRADIUS,'arrowscale':EVECDEFAULT_ARROWSCALE, 'drawbackground':EVECDEFAULT_DRAWBACKGROUND}

FORCEDEFAULT_TIPLENGTH: float = 0.4
FORCEDEFAULT_TIPRADIUS: float = 0.25
FORCEDEFAULT_ARROWSCALE: float = 10000.0
FORCEDEFAULT_DRAWBACKGROUND: bool = True
FORCEDEFAULT_SETTINGS={'tiplength':FORCEDEFAULT_TIPLENGTH,'tipradius':FORCEDEFAULT_TIPRADIUS,'arrowscale':FORCEDEFAULT_ARROWSCALE, 'drawbackground':FORCEDEFAULT_DRAWBACKGROUND}
