r"""
Constants for setting up the main CSPS class. Contains:
- Necessary keys and their data types in settings
"""
# Necessary keys =======================================================================================================
KEYS = {'sps_attempts': int, 'test_convex_initial': bool, 'algo_displace': int, 'i_cycle_after_displace': bool,
        'disp2_superpos_modenr_low': int, 'disp2_superpos_modenr_high': int, 'disp2_eigalgo': int,
        'disp2_stepscale': float,
        'algo_escape': int, 'i_start_after_displace': bool, 'i_cycle_after_escape': bool, 'esc1_modenr_low': int,
        'esc1_modenr_high': int, 'algo_converge': int, 'i_start_after_escape': bool, 'i_cycle_after_converge': bool,
        'algo_escape_again': int, 'i_start_after_converge': bool, 'i_write_initial_eigenvecs': [bool, int]}
